################################# LIBRARIES ###################################
# Read Excel data
library("readxl")
# Dot product
library("geometry")

################################# SETTINGS ###################################
# Read time series from xlsx file (see example file for formatting
# specifics)
X <- read_excel("EXAMPLE DATA/muc_annual_daily_maximum_rainfall.xlsx")
# Start time, end time, time interval based on time series
t_start <- 1
t_end <- length(X)
dt <- 1
T <- length(seq(t_start, t_end, dt))
# How far into the future to predict
forecast_horizon = 10
# Cube length R of C_R
R <- 10
# Penalty weight
gamma <- 0.3
# Threshold for termination of gradient descent
delta <- 0.0001
# Choose between variants with (0) or without bootstrapping (1)
variant <- 0
# Choose if cleaning of learning samples should be enabled (1) or not (0)
clean <- 1

############################# FUNCTIONS ####################################

check_for_cleanness <- function(j, learning_samples, observed_values, X_tilde, X_tilde_hat, lambda, F, f, gamma, variant){
    # For a given lambda, this function checks if the assumptions for clean
    # learning samples are given

    # First get gradients for all learning samples
    nabla_Q <- matrix(0, dim(learning_samples)[2], dim(learning_samples)[1])
    for (i in seq(1, dim(nabla_Q)[2], 1)){
        if (!variant){
            gradient_list = get_gradient(i, learning_samples, observed_values, X_tilde, X_tilde_hat, lambda, F, f, gamma)
            nabla_Q[, i] <- gradient_list[[1]]
        } else {
            gradient_list = get_gradient_variant(i, learning_samples, observed_values, lambda, F, f, gamma)
            nabla_Q[, i] <- gradient_list[[1]]
        }
    }

    # Check for 1st condition (gradient of Q_j is non-zero)
    if (norm(nabla_Q[, j], type = "2") == 0){
        return(0)
    }

    # Check for 2nd condition (scalar products of Q_j and Q_i are zero)
    for (i in 1:dim(nabla_Q)[2]){
        if (i == j){
            next
        }
        if (dot(nabla_Q[, j], nabla_Q[, i], d = TRUE) != 0){
            return(0)
        }
    }
    return(1)
}

extrapolation <- function(X, t_start, t_end, dt, R, alpha, loc, shape, gamma, delta, forecast_horizon, variant, clean){
    # Performs the extrapolation

    # Fr<e9>chet CDF and PDF
    F <- function(x){
        return(exp(-((x-loc)/shape)^(-alpha)))
    }

    f <- function(x){
        (alpha/shape)*((x-loc)/shape)^(-alpha-1)*exp(-((x-loc)/shape)^(-alpha))
    }
    
    time <- seq(from = t_start, to = t_end, by = dt)
    T <- length(time)

    # Initialize forecast sample
    forecast_sample <- X[seq(T-R, T, 1)]

    # Get learning samples and the values corresponding to forecast
    # horizon
    list_learning_samples = get_learning_samples(X, T, R, length(forecast_sample), forecast_horizon)
    learning_samples = list_learning_samples[[1]]
    observed_values = list_learning_samples[[2]]

    # Initial guess, tunable, must be same size as forecast sample
    lambda0 <- seq(from=0.7, to=1, by=0.3/R)
    if (length(lambda0) != length(forecast_sample)){
        stop("Initial guess lambda0 must be the same size as forecast sample!")
    }
    # Calculate lambda and max-stable predictor
    lambda_end <- gradient_descent(lambda0, learning_samples, observed_values, F, f, gamma, delta, variant, clean)
    X_hat <- get_extrapolator(forecast_sample, lambda_end)

    return(X_hat)
}

get_extrapolator <- function(observations, lambda){
    # Calculates the max-stable predictor

    if (length(observations) != length(lambda)){
        disp('get_extrapolator: Invalid input. 0 returned.')
        x <- 0
    } else {
        x <- max(observations*lambda)
    }
    return(x)
}

get_gradient_variant <- function(idx, learning_samples, observed_values, lambda, F, f, gamma){
    # Computes the gradient of Q_j of alternative formulation of Phi
    N <- dim(learning_samples)[1]

    # X_h is the observation that corresponds to forecast
    X_h <- observed_values[idx]
    # Initialize learning sample
    sub_sample <- learning_samples[idx, ]
    # Calculate predictor
    X_hat_h <- max(sub_sample*lambda)

    # Define Gradient
    dQ <- rep(1, length(lambda))
    non_differentiable <- 0

    for (j in seq(from = 1, to = length(lambda), by = 1)){
        # Always check if Phi is differentiable. This is the case if there is
        # no turnover/kink in the maximum functions
        X_hat_sum_j <- 0
        X_hat_sum_i <- 0
        if(idx>1){
            for (i in seq(1,idx-1,1)){
                sub_sample_i <- learning_samples[i,]
                X_hat_sum_j <- X_hat_sum_j + f(lambda[j]*sub_sample[j])*sub_sample[j]*(max(lambda*sub_sample_i) < max(lambda*sub_sample))
                X_hat_sum_i <- X_hat_sum_i + f(lambda[j]*sub_sample_i[j])*sub_sample_i[j]*(max(lambda*sub_sample_i) > max(lambda*sub_sample))*(lambda[j]*sub_sample_i[j] == X_hat_h)
                if (max(lambda*sub_sample_i) == max(lambda*sub_sample)){
                    # Kink occurs, Phi is not differentiable here
                    non_differentiable <- 1
                }
            }
        }
        # print(observed_values)
        # print(X_h)
        # print(lambda)
        # print(sub_sample)
        if ( X_h==lambda[j]*sub_sample[j] || sum((sub_sample*lambda)==X_hat_h)>1){
            # Kink occurs, Phi is not differentiable here
            dQ[j] <- 0
            non_differentiable <- 1
        } else {
            dQ[j] <- (f(lambda[j]*sub_sample[j])*sub_sample[j]*(2*(X_h<lambda[j]*sub_sample[j]) - 1 + gamma/N - 2*gamma*F(lambda[j]*sub_sample[j])) + (2*gamma/N)*X_hat_sum_j)*(lambda[j]*sub_sample[j] == X_hat_h) + (2*gamma/N)*X_hat_sum_i
        }
    }
    
    return(list(dQ, non_differentiable))
}

get_gradient <- function(idx, learning_samples, observed_values, X_tilde, X_tilde_hat, lambda, F, f, gamma){
 # Computes the gradient of Q_j of Phi

    # X_h is the observation that corresponds to forecast within the
    # next learning sample position-wise
    X_h <- observed_values[idx]
    # Initalize learning sample
    sub_sample <- learning_samples[idx, ]
    # Calculate predictor
    X_hat_h <- get_extrapolator (sub_sample, lambda)

    # Define corresponding X_tilde values (from bootstrap samples)
    X_tilde_h <- X_tilde[idx, ]
    X_tilde_hat_h <- X_tilde_hat[idx]

    # Define Gradient
    dQ <- rep(1,length(lambda))
    non_differentiable <- 0
    for (j in 1:length(lambda)){
        # Check if Phi is differentiable. This is the case if there is
        # no turnover/kink in the maximum functions
        if (X_tilde_hat_h == X_hat_h || X_h==lambda[j]*sub_sample[j] || sum((sub_sample*lambda)==X_hat_h)>1){
            # Kink occurs, Phi is not differentiable here
            non_differentiable <- 1
        } else {
            dq <- (2*(X_h<lambda[j]*sub_sample[j]) - 1 - gamma*(X_tilde_hat_h<X_hat_h)+2*gamma*F(lambda[j]*sub_sample[j]))*f(lambda[j]*sub_sample[j])*sub_sample[j]*(lambda[j]*sub_sample[j] == X_hat_h)

            dp <- gamma*(X_tilde_hat_h>X_hat_h)*f(X_tilde_hat_h)*X_tilde_h[j]*(lambda[j]*X_tilde_h[j] == X_tilde_hat_h)

            dQ[j] <- dq-dp
        }
    }
    return(list(dQ, non_differentiable))
}

get_learning_samples <- function(X, T, R, len, forecast_horizon){
    # Creates and returns learning samples

    # Amount of learning samples; by default takes as much available data
    # as possible
    ls_count <- floor((T-R-1)/R)

    # Store learning samples row-wise in a matrix and the observed values
    # corresponding to forecast horizon in a vector
    learning_samples <- matrix(0, ls_count, len)
    observed_values <- rep(0, ls_count)
    # Locations/time stamps of the learning samples
    loc <- seq(T-2*R, 1, -R)

    # Initialize learning samples with values from X
    for (h in seq(1,ls_count,1)){
        learning_samples[h,] <- t(X[seq(loc[h], loc[h]+len-1, 1)])
        observed_values[h] <- X[loc[h]+len-1+forecast_horizon][[1]]
    }
    
    return(list(learning_samples, observed_values))
}

gradient_descent <- function(lambda_start, learning_samples, observed_values, F, f, gamma, delta, variant, clean){
    # Performs stochastic gradient descent for the function Phi

    # Initial step size for harmonic series (1/n) is 1/1
    step_size_divisor <- 1

    # Initialize final lambda
    lambda_end <- lambda_start
    # lambda variable to store values of each iteration
    lambda <- lambda_start
    # lambda of previous iteration to check the termination condition
    lambda_prev <- 0
    # Dimensions of learning_sample matrix
    nrow = dim(learning_samples)[1]
    ncol = dim(learning_samples)[2]

    # Generate Bootstrap samples
    X_tilde <- matrix(0, nrow, ncol)
    X_tilde_hat <- rep(0, nrow)
    Y <- rep(0, nrow)
    for (i in seq(1, dim(X_tilde)[1], 1)){
        for (j in seq(1, dim(X_tilde)[2], 1)){
            X_tilde[i,j] <- learning_samples[sample(1:nrow, 1), sample(1:ncol, 1)]
        }
        X_tilde_hat[i] <- get_extrapolator(X_tilde[i,], lambda)
        Y[i] <- F(X_tilde_hat[i])
    }

    while (norm(as.matrix(lambda-lambda_prev), type = "2") > delta){

        # Choose learning sample randomly
        hj <- sample(1:nrow, 1)

        # Compute Gradient
        if (!variant){
            gradient_list <- get_gradient(hj, learning_samples, observed_values, X_tilde, X_tilde_hat, lambda, F, f, gamma)
        } else {
            gradient_list <- get_gradient_variant(hj, learning_samples, observed_values, lambda, F, f, gamma)
        }

        dQ = gradient_list[1][[1]]
        non_differentiable = gradient_list[2][[1]]

        # If clean is enabled then check for the cleanness of the
        # gradients, else just for non differentiablity
        if (clean){
            if (non_differentiable || check_for_cleanness(hj, learning_samples, observed_values, X_tilde, X_tilde_hat, lambda, F, f, gamma, variant)){
                next
            }
        } else {
            if (non_differentiable){
                next
            }
        }

        # Stochastic Gradient Descent
        lambda_prev <- lambda
        lambda <- lambda - (1/step_size_divisor)*dQ

        # Adjust step size, according to harmonic series (1/n)
        step_size_divisor <- step_size_divisor+1

        # Check if new lambda makes any improvement of the solution. If so,
        # store it.
        if (!variant){
            if (phi(lambda_end, learning_samples, observed_values, Y, F, gamma) > phi(lambda, learning_samples, observed_values, Y, F, gamma)){
                lambda_end <- lambda
            }
        } else {
            if (phi_variant(lambda_end, learning_samples, observed_values, F, gamma) > phi_variant(lambda, learning_samples, observed_values, F, gamma)){
                lambda_end <- lambda
            }
        }
    }
    return(lambda_end)
}

phi_variant <- function(lambda, learning_samples, observed_values, F, gamma){
    # Calculates the altered version of the function Phi

    y <- 0
    N <- dim(learning_samples)[1]
    for (j in seq(1,N,1)){
        # X_j is the observation that corresponds to forecast within the
        # next learning sample position-wise
        X_j <- observed_values[j]
        sub_sample <- learning_samples[j, ]
        X_hat_j <- max(sub_sample*lambda)

        X_hat_sum <- 0
        if(j>1){
            for (i in seq(1, j-1, 1)){
            sub_sample_i <- learning_samples[i, ]
            X_hat_sum <- X_hat_sum + F(max(max(lambda*sub_sample), max(lambda*sub_sample_i)))
            }
            y <- y + 2*F(max(X_j, X_hat_j)) - F(X_hat_j) - gamma*F(X_hat_j)^2+ (gamma/N)*(F(X_hat_j)+2*X_hat_sum)
        }
    }
    return(y)
}

phi <- function(lambda, learning_samples, observed_values, Y, F, gamma){
    # Calculates the function Phi
    y <- 0
    for (hj in seq(1, dim(learning_samples)[1], 1)){
        # X_hj is the observation that corresponds to forecast within the
        # next learning sample position-wise
        X_hj <- observed_values[hj]
        sub_sample <- learning_samples[j, ]
        X_hat_hj <- max(sub_sample*lambda)

        y <- y + 2*F(max(X_hj, X_hat_hj)) - F(X_hat_hj) + gamma*((1/3) - max(F(X_hat_hj), Y[hj]) + F(X_hat_hj)^2)
    }
    return(y)
}

############################# EXTRAPOLATION ################################

# Estimated Frechet parameters of example data
alpha_est = 7.7551
loc_est = -545.0173 
shape_est = 959.8184

# Vector to store forecasts
X_hat = rep(0, forecast_horizon)
# Make forecasts
for (j in seq(1, forecast_horizon, 1)){
   X_hat[j] <- extrapolation(X, t_start, t_end, dt, R, alpha_est, loc_est, shape_est, gamma, delta, j, variant, clean)
}

print("Forecast:")
print(X_hat)








